import os
import random
import re
import shutil
import time

import selenium.common
import selenium.webdriver
from selenium.common import WebDriverException
from selenium.webdriver.common.by import By
from selenium.webdriver.firefox.service import Service
from selenium.webdriver.firefox.webdriver import WebDriver
from selenium.webdriver.support import expected_conditions
from selenium.webdriver.support.ui import WebDriverWait

from modulitiz.browser.AbstractModuloBrowser import AbstractModuloBrowser
from modulitiz.browser.TextToBePresentInElements import TextToBePresentInElements
from modulitiz.browser.constants.BrowserConstants import BrowserConstants
from modulitiz.browser.constants.JsScripts import JsScripts
from modulitiz.browser.enums.BrowserAddonExtensionEnum import BrowserAddonExtensionEnum
from modulitiz.browser.enums.BrowserProxyTypeEnum import BrowserProxyTypeEnum
from modulitiz_micro.ModuloStringhe import ModuloStringhe
from modulitiz_micro.eccezioni.EccezioneRuntime import EccezioneRuntime
from modulitiz_micro.eccezioni.http.EccezioneHttpGeneric import EccezioneHttpGeneric
from modulitiz_micro.files.ModuloFiles import ModuloFiles
from modulitiz_micro.sistema.EnvVarsEnum import EnvVarsEnum
from modulitiz_micro.sistema.ModuloEnvVars import ModuloEnvVars
from modulitiz_micro.sistema.ModuloSystem import ModuloSystem


class ModuloBrowser(AbstractModuloBrowser):
	
	def __init__(self,cartellaGeckodriver:str):
		super().__init__()
		if ModuloSystem.isWindows():
			fullpathGeckodriver=ModuloFiles.pathJoin(cartellaGeckodriver,"geckodriver_x")
			if ModuloSystem.isSystem64bit() is True:
				fullpathGeckodriver+="64"
			else:
				fullpathGeckodriver+="32"
			fullpathGeckodriver+=".exe"
		else:
			fullpathGeckodriver="/snap/bin/firefox.geckodriver"
		self.fullpathGeckodriver=fullpathGeckodriver
		self.pathTemp=ModuloFiles.pathJoin(ModuloSystem.getTempFolder(),"firefox_selenium")
		self.percorsoCartellaProfilo=None
		self.driver:WebDriver|None=None
	
	def populate(self,percorsoCartellaFirefox:str):
		os.makedirs(self.pathTemp,exist_ok=True)
		ModuloEnvVars.add(EnvVarsEnum.PATH,percorsoCartellaFirefox,False)
	
	# =================================================================================================================
	# ======================================== INIZIALIZZAZIONE E CHIUSURA ============================================
	# =================================================================================================================
	def createDriver(self,nomefileLog:str|None,profiloBrowser:selenium.webdriver.FirefoxProfile,hideWindow:bool,privateMode:bool):
		if nomefileLog is None:
			nomefileLog=ModuloSystem.getFolderNull()
		options=self.createOptions(hideWindow,privateMode)
		options.profile=profiloBrowser
		# creo il driver
		self.driver=selenium.webdriver.Firefox(options,Service(self.fullpathGeckodriver,log_output=nomefileLog))
		# get cartella effettiva profilo
		self.percorsoCartellaProfilo=profiloBrowser._profile_dir
	
	@staticmethod
	def createOptions(hideWindow:bool,privateMode:bool)->selenium.webdriver.FirefoxOptions:
		options=selenium.webdriver.FirefoxOptions()
		options.headless=hideWindow
		if privateMode is True:
			options.add_argument("-private-window")
		return options
	
	def createProfile(self,cartellaProfilo:str|None,permessoImmagini:int|None,noAudio:bool,grantPermissions:bool,openNewTab:bool,
			cartellaDownload:str|None,socksProxy:tuple|None,protezioneAntiTracciamento:bool,userAgent:str|None)->selenium.webdriver.FirefoxProfile:
		if cartellaProfilo is None:
			profilo=selenium.webdriver.FirefoxProfile()
		else:
			profilo=selenium.webdriver.FirefoxProfile(cartellaProfilo)
		# parametric options
		if permessoImmagini is not None:
			profilo.set_preference("permissions.default.image",permessoImmagini)
		if noAudio is True:
			profilo.set_preference("media.volume_scale", "0.0")
		if grantPermissions is True:
			profilo.set_preference('media.navigator.permission.disabled',True)
		if openNewTab is True:
			profilo.set_preference("browser.link.open_newwindow",3)
			profilo.set_preference("browser.link.open_newwindow.restriction",0)
			profilo.set_preference("browser.link.open_newwindow.override.external",3)
		if cartellaDownload is not None:
			profilo.set_preference("browser.preferences.instantApply",True)
			profilo.set_preference("browser.helperApps.neverAsk.saveToDisk",BrowserConstants.NEVER_ASK_SAVE_FILES)
			profilo.set_preference("browser.helperApps.alwaysAsk.force",False)
			profilo.set_preference("browser.helperApps.alwaysAsk.focusWhenStarting",False)
			profilo.set_preference('browser.download.manager.showWhenStarting',False)
			profilo.set_preference('browser.download.manager.useWindow',False)
			profilo.set_preference("browser.download.panel.shown",False)
			profilo.set_preference('browser.download.folderList',2)
			profilo.set_preference('browser.download.dir',cartellaDownload)
		if socksProxy is not None:
			socksProxyIp,socksProxyPorta=socksProxy
			if socksProxyIp is None:
				socksProxyIp="127.0.0.1"
			profilo.set_preference('network.proxy.type',BrowserProxyTypeEnum.MANUAL)
			profilo.set_preference('network.proxy.socks',socksProxyIp)
			profilo.set_preference('network.proxy.socks_port',socksProxyPorta)
			profilo.set_preference('network.proxy.socks_version',5)
			profilo.set_preference('network.proxy.socks_remote_dns',True)
		if protezioneAntiTracciamento is not None:
			profilo.set_preference('privacy.trackingprotection.enabled',protezioneAntiTracciamento)
			profilo.set_preference('privacy.trackingprotection.socialtracking.enabled',protezioneAntiTracciamento)
		if userAgent is not None:
			profilo.set_preference("general.useragent.override",userAgent)
		return profilo
	
	def closeAll(self):
		if self.driver is None:
			return
		# delete temp dir
		if os.path.exists(self.pathTemp):
			shutil.rmtree(self.pathTemp)
		# delete profile dir
		if os.path.exists(self.percorsoCartellaProfilo):
			shutil.rmtree(self.percorsoCartellaProfilo)
		# close driver
		self.driver.quit()
		self.driver=None
	
	# =================================================================================================================
	# ================================================= GESTIONE HTML =================================================
	# =================================================================================================================
	
	def switchIframeByCss(self,cssSelector:str,waitUntilTimeout: int|None):
		"""
		:param cssSelector: Selettore css per identificare l'elemento
		:param waitUntilTimeout: aspetta fino a n secondi se l'elemento non e' ancora disponibile, se si e' sicuri che esista sempre mettere None
		"""
		return WebDriverWait(self.driver,waitUntilTimeout).until(
			expected_conditions.frame_to_be_available_and_switch_to_it((By.CSS_SELECTOR,cssSelector))
		)
	
	def switchToDefaultFrame(self):
		"""
		Ritorna al frame (o iframe) precedente.
		"""
		self.driver.switch_to.default_content()
	
	def getElementsByCss(self,cssSelector:str,waitUntilTimeout: int|None,textInside: str|None):
		"""
		:param cssSelector: Selettore css per identificare gli elementi
		:param waitUntilTimeout: aspetta fino a n secondi se gli elementi non sono ancora disponibili, se si è sicuri che esistano sempre mettere None
		:param textInside: considera anche il testo all'interno dei tag
		"""
		if textInside is not None:
			return WebDriverWait(self.driver,waitUntilTimeout).until(
				TextToBePresentInElements((By.CSS_SELECTOR,cssSelector),textInside)
			)
		if waitUntilTimeout is None:
			return self.driver.find_elements(By.CSS_SELECTOR,cssSelector)
		try:
			return WebDriverWait(self.driver,waitUntilTimeout).until(
				expected_conditions.visibility_of_element_located((By.CSS_SELECTOR,cssSelector))
			)
		except selenium.common.exceptions.TimeoutException:
			return None
	
	def goToUrl(self,url:str,maxRetries:int=3,secPausaMin:int=5,secPausaMax:int=15):
		while maxRetries>0:
			try:
				self.driver.get(url)
				maxRetries=0
			except EccezioneHttpGeneric:
				time.sleep(random.randint(secPausaMin,secPausaMax))
				maxRetries-=1
	
	def getTabTitle(self)->str:
		windowHandle=self.driver.window_handles
		# cambio la scheda
		for windowId in self.driver.window_handles:
			if windowId!=windowHandle:
				self.driver.switch_to.window(windowId)
		# ricavo il titolo
		time.sleep(0.2)
		nomeScheda=self.driver.title
		return nomeScheda
	
	# =================================================================================================================
	# ============================================= PROPRIETA' FINESTRA ===============================================
	# =================================================================================================================
	def get_dimensione_finestra(self)->tuple:
		size=self.driver.get_window_size()
		larghezza=size['width']
		altezza=size['height']
		return larghezza,altezza
	
	def set_dimensione_finestra(self,larghezza:int,altezza:int):
		# massimizzo la finestra per sapere le dimensioni massime
		larghezzaPrec,altezzaPrec=self.get_dimensione_finestra()
		self.maximizeWindow()
		larghezza_max,altezza_max=self.get_dimensione_finestra()
		# imposto la dimensione desiderata se non supera la dimensione massima
		if larghezza>larghezza_max or altezza>altezza_max:
			self.driver.set_window_size(larghezzaPrec,altezzaPrec)
			return
		self.driver.set_window_size(larghezza,altezza)
	
	def set_dimensione_finestra_random(self)->tuple:
		self.driver.maximize_window()
		larghezzaMax,altezzaMax=self.get_dimensione_finestra()
		larghezza=random.randint(800,larghezzaMax)
		altezza=random.randint(500,altezzaMax)
		self.set_dimensione_finestra(larghezza,altezza)
		return larghezza,altezza
	
	# =================================================================================================================
	# ================================================== ESTENSIONI ===================================================
	# =================================================================================================================
	def aggiungi_estensioni(self,cartella:str,nomefiles:list):
		# per ogni file...
		for nomefile in nomefiles:
			#...controllo che abbia la giusta estensione
			if nomefile.endswith(BrowserAddonExtensionEnum.FIREFOX) is False:
				raise EccezioneRuntime("Il file "+nomefile+" non e' riconosciuto come estensione.")
			# installo l'estensione
			percorsoNomefile=ModuloFiles.pathJoin(cartella,nomefile)
			self.driver.install_addon(percorsoNomefile,temporary=True)
	
	def get_estensione_baseurl_da_nomefile(self,nomeEstensione:str)-> str|None:
		self.apri_nuova_scheda()
		self.driver.get("about:memory")
		bottoneMisura=self.getElementsByCss('div.opsRow button#measureButton',10,None)[0]
		bottoneMisura.click()
		
		nomeEstensione="id="+nomeEstensione
		elemento=self.getElementsByCss('span[id*="Process:extensions"]+span.kids > span.mrName', 10, nomeEstensione)[0]
		stringaEstensione=elemento.text
		ricerca=re.search(r""", baseURL=(.+)\)""",stringaEstensione)
		if ricerca is None:
			return None
		self.chiudi_scheda_corrente()
		return ricerca.group(1)
	
	# =================================================================================================================
	# ================================================== ALERT ========================================================
	# =================================================================================================================
	def gestione_alert(self):
		WebDriverWait(self.driver,10).until(
			expected_conditions.alert_is_present()
		)
		alertWindow=self.driver.switch_to.alert
		alertWindow.accept()
	
	# =================================================================================================================
	# ============================================== GESTIONE SCHEDE ==================================================
	# =================================================================================================================
	def retrieveOpenedTabs(self)->int:
		"""
		Recupera il numero di schede aperte.
		"""
		try:
			return len(self.driver.window_handles)
		except WebDriverException:
			return 0
		
	def apri_nuova_scheda(self)->int:
		numSchedeApertePrima=len(self.driver.window_handles)
		self.driver.execute_script(JsScripts.OPEN_NEW_WINDOW)
		numSchedeAperteDopo=len(self.driver.window_handles)
		if numSchedeApertePrima==numSchedeAperteDopo:
			raise EccezioneRuntime("Non sono riuscito ad aprire una nuova scheda")
		self.driver.switch_to.window(self.driver.window_handles[-1])
		return numSchedeAperteDopo-1
	
	def chiudi_scheda_corrente(self)->int:
		self.driver.close()
		self.driver.switch_to.window(self.driver.window_handles[0])
		return 0
	
	def chiudi_scheda_corrente_se_non_vuota(self):
		self.waitUntilPageFullyLoaded()
		if ModuloStringhe.isEmpty(self.getTabTitle()):
			return
		self.chiudi_scheda_corrente()
